from typing import Optional
from decimal import Decimal

from validator_collection import validators, checkers

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.options.exporting.csv import ExportingCSV
from highcharts_core.options.exporting.pdf_font import PDFFontOptions
from highcharts_core.utility_classes.menus import MenuObject
from highcharts_core.utility_classes.buttons import ContextButtonConfiguration, \
    ExportingButtons
from highcharts_core.utility_classes.javascript_functions import CallbackFunction

default_context_button = ExportingButtons()
default_context_button['contextButton'] = ContextButtonConfiguration()


class ExportingAccessibilityOptions(HighchartsMeta):
    """Accessibility options for the exporting menu."""

    def __init_(self, **kwargs):
        self._enabled = None

        self.enabled = kwargs.get('enabled', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'exporting.accessibility'

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enables accessibility support for the export menu. Defaults to
        ``True``.

        :returns: Flag indicating whether accessibility support is enabled for the
          export menu.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'enabled': as_dict.get('enabled', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'enabled': self.enabled
        }


class Exporting(HighchartsMeta):
    """Options to configure the export functionality enabled for the chart."""

    def __init__(self, **kwargs):
        self._accessibility = None
        self._allow_html = None
        self._buttons = None
        self._chart_options = None
        self._csv = None
        self._enabled = None
        self._error = None
        self._fallback_to_export_server = None
        self._filename = None
        self._form_attributes = None
        self._lib_url = None
        self._menu_item_definitions = None
        self._pdf_font = None
        self._print_max_width = None
        self._scale = None
        self._show_table = None
        self._source_height = None
        self._source_width = None
        self._table_caption = None
        self._type = None
        self._url = None
        self._use_multi_level_headers = None
        self._use_rowspan_headers = None
        self._width = None

        self.accessibility = kwargs.get('accessibility', None)
        self.allow_html = kwargs.get('allow_html', None)
        self.buttons = kwargs.get('buttons', default_context_button)
        self.chart_options = kwargs.get('chart_options', None)
        self.csv = kwargs.get('csv', None)
        self.enabled = kwargs.get('enabled', None)
        self.error = kwargs.get('error', None)
        self.fallback_to_export_server = kwargs.get('fallback_to_export_server', None)
        self.filename = kwargs.get('filename', None)
        self.form_attributes = kwargs.get('form_attributes', None)
        self.lib_url = kwargs.get('lib_url', None)
        self.menu_item_definitions = kwargs.get('menu_item_definitions', None)
        self.pdf_font = kwargs.get('pdf_font', None)
        self.print_max_width = kwargs.get('print_max_width', None)
        self.scale = kwargs.get('scale', None)
        self.show_table = kwargs.get('show_table', None)
        self.source_height = kwargs.get('source_height', None)
        self.source_width = kwargs.get('source_width', None)
        self.table_caption = kwargs.get('table_caption', None)
        self.type = kwargs.get('type', None)
        self.url = kwargs.get('url', None)
        self.use_multi_level_headers = kwargs.get('use_multi_level_headers', None)
        self.use_rowspan_headers = kwargs.get('use_rowspan_headers', None)
        self.width = kwargs.get('width', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'exporting'

    @property
    def accessibility(self) -> Optional[ExportingAccessibilityOptions]:
        """Accessibility options for the exporting menu.

        :rtype: :class:`ExportingAccessibilityOptions` or :obj:`None <python:None>`
        """
        return self._accessibility

    @accessibility.setter
    @class_sensitive(ExportingAccessibilityOptions)
    def accessibility(self, value):
        self._accessibility = value

    @property
    def allow_html(self) -> Optional[bool]:
        """If ``True``, allows HTML inside the chart (added using
        ``.use_html`` properties present on various chart components) to be added 
        directly to the exported image. This allows you to preserve complicated HTML 
        structures like tables or bi-directional text in exported charts.

        Defaults to ``False``.

        .. warning::

          This setting is **EXPERIMENTAL**.

          The HTML is rendered in a ``foreignObject`` tag in the generated SVG. The
          official export server is based on PhantomJS, which supports this, but other 
          SVG clients, like Batik, do not support it. This also applies to downloaded 
          SVG that you want to open in a desktop client.

        :returns: Flag indicating whether to allow HTML in the exported image.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_html

    @allow_html.setter
    def allow_html(self, value):
        if value is None:
            self._allow_html = None
        else:
            self._allow_html = bool(value)

    @property
    def buttons(self) -> Optional[ExportingButtons]:
        """Options for the export related buttons: print and export.

        .. note::

          In addition to the default buttons listed above, custom buttons can be added.
          
        .. warning::
        
          The ``.buttons`` property accepts an 
          :class:`ExportingButtons <highcharts_core.utility_classes.buttons.ExportingButtons>`
          instance as its value. This object is a descendent of the special :class:`JavaScriptDict <highcharts_core.metaclasses.JavaScriptDict>`
          which by default initially contains a ``'context

        :rtype: :class:`ExportingButtons`
        """
        return self._buttons

    @buttons.setter
    @class_sensitive(ExportingButtons)
    def buttons(self, value):
        self._buttons = value

    @property
    def chart_options(self):
        """Additional chart options to be merged into the chart before exporting to an
        image format. This does not apply to printing the chart via the export menu.

        For example, a common use case is to add data labels to improve readability of the
        exported chart, or to add a printer-friendly color scheme to exported PDFs.

        .. warning::

          To avoid a circular import error, this property **REQUIRES** that you supply a
          value that is an :class:`Options` instance (e.g. :class:`HighchartsOptions`,
          :class:`HighchartsStockOptions`, :class:`HighchartsMapsOptions`, etc.). Unlike
          other Highcharts for Python properties, it does **not** accept
          :class:`dict <python:dict>` or JSON (:class:`str <python:dict>`) values.

          Please be sure to either supply it a valid :class:`Options` instance, or the
          value of :obj:`None <python:None>`.

        :rtype: :class:`Options` or :obj:`None <python:None>`
        """
        return self._chart_options

    @chart_options.setter
    def chart_options(self, value):
        if not value:
            self._chart_options = None
        elif not checkers.is_type(value, 'Options'):
            raise errors.InstanceNeededError(f'The Exporting.chart_options property is '
                                             f'one of the few properties in Highcharts '
                                             f'for Python that REQUIRES a Highcharts for '
                                             f'Python instance as its value (or None). '
                                             f'Specifically, you should supply an Options'
                                             f' instance to it, rather than a dict or a '
                                             f'string. The value you supplied was: '
                                             f'{value.__class__.__name}')
        else:
            self._chart_options = value

    @property
    def csv(self) -> Optional[ExportingCSV]:
        """Options for exporting data to CSV or Microsoft Excel, or displaying the data in
        a HTML table or a JavaScript structure.

        This module adds data export options to the export menu and provides JavaScript
        functions like ``Chart.getCSV()``, ``Chart.getTable()``, ``Chart.getDataRows()``,
        and ``Chart.viewData()``.

        .. warning::

          The XLS converter is limited and only creates a HTML string that is passed for
          download, which works but creates a warning before opening. The workaround for
          this is to use a third party XLSX converter.

        :returns: Configuration for exporting data to CSV or Microsoft Excel.
        :rtype: :class:`ExportingCSV` or :obj:`None <python:None>`
        """
        return self._csv

    @csv.setter
    @class_sensitive(ExportingCSV)
    def csv(self, value):
        self._csv = value

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, displays the export context button and allows for exporting the
        chart. If ``False``, the context button will be hidden but JavaScript export API
        methods will still be available.

        Defaults to ``True``.

        :returns: Flag indicating whether the export menu is displayed on the chart.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def error(self) -> Optional[CallbackFunction]:
        """JavaScript function that is called if the offline-exporting module fails to
        export a chart on the client side, and :meth:`Exporting.fallback_to_export_server`
        is disabled.

        If :obj:`None <python:None>`, a JavaScript exception is thrown instead. The
        JavaScript function receives two parameters, the exporting options, and the error
        from the module.

        .. seealso::

          * :meth:`Exporting.fallback_to_export_server`

        :returns: JavaScript function code
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._error

    @error.setter
    @class_sensitive(CallbackFunction)
    def error(self, value):
        self._error = value

    @property
    def fallback_to_export_server(self) -> Optional[bool]:
        """If ``True``, falls back to the export server if the offline-exporting module is
        unable to export the chart on the client side. Defaults to ``True``.

        This happens for certain browsers, and certain features (e.g.
        :meth:`Exporting.allow_html`), depending on the image type exporting to.

        .. hint::

          For very complex charts, it is possible that export can fail in browsers that
          don't support Blob objects, due to data URL length limits. It is recommended to
          define the :meth:`Exporting.error` handler if disabling fallback, in order to
          notify users in case export fails.

        :returns: Flag indicating whether to fall back to the export server if chart
          export fails.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._fallback_to_export_server

    @fallback_to_export_server.setter
    def fallback_to_export_server(self, value):
        if value is None:
            self._fallback_to_export_server = None
        else:
            self._fallback_to_export_server = bool(value)

    @property
    def filename(self) -> Optional[str]:
        """The filename (without file type extension) to use for the exported chart.
        Defaults to ``'{constants.DEFAULT_EXPORTING_FILENAME}'``.

        :rtype: :class:`str` or :obj:`None <python:None>`
        """
        return self._filename

    @filename.setter
    def filename(self, value):
        self._filename = validators.string(value, allow_empty = True)

    @property
    def form_attributes(self) -> Optional[dict]:
        """An object containing additional key value data for the POST form that sends the
        SVG to the export server.

        For example, a ``target`` can be set to make sure the generated image is received
        in another frame, or a custom ``enctype`` or ``encoding`` can be set.

        :returns: Additional form attributes to supply to the export server.
        :rtype: :class:`dict <python:dict>` or :obj:`None <python:None>`
        """
        return self._form_attributes

    @form_attributes.setter
    def form_attributes(self, value):
        self._form_attributes = validators.dict(value, allow_empty = True)

    @property
    def lib_url(self) -> Optional[str]:
        """Path where Highcharts will look for export module dependencies to load on
        demand if they don't already exist on window.

        Should currently point to location of the
        `CanVG <https://github.com/canvg/canvg>`_ library,
        `jsPDF <https://github.com/parallax/jsPDF>`_ and
        `svg2pdf.js <https://github.com/yWorks/svg2pdf.js>`_, which are all required for
        client side export in certain browsers.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._lib_url

    @lib_url.setter
    def lib_url(self, value):
        self._lib_url = validators.url(value, allow_empty = True)

    @property
    def menu_item_definitions(self) -> Optional[MenuObject]:
        """An object consisting of definitions for the menu items in the context menu.

        Each key value pair has a key that is referenced in the ``menu_items`` setting,
        and a value, which is an object with the following properties:

          * ``onclick``: The click handler for the menu item
          * ``text``: The text for the menu item
          * ``textKey``: If internationalization is required, the key to a language 
            string

        .. note::

          Custom text for ``"exitFullScreen"`` can be set only in ``language`` options 
          (it is not a separate button).

        Defaults to:

        .. code-block:: python

          {
              "viewFullscreen": {},
              "printChart": {},
              "separator": {},
              "downloadPNG": {},
              "downloadJPEG": {},
              "downloadPDF": {},
              "downloadSVG": {}
          }

        :returns: Definitions for menu items in the Exporting context menu.
        :rtype: :class:`MenuObject` or :obj:`None <python:None>`
        """
        return self._menu_item_definitions

    @menu_item_definitions.setter
    @class_sensitive(MenuObject)
    def menu_item_definitions(self, value):
        self._menu_item_definitions = value

    @property
    def pdf_font(self) -> Optional[PDFFontOptions]:
        """Settings for a custom font for the exported PDF, when using the
        ``offline-exporting`` module.

        This is used for languages containing non-ASCII characters, like Chinese, Russian,
        Japanese etc.

        As described in the
        `jsPDF docs <https://github.com/parallax/jsPDF#use-of-unicode-characters--utf-8>`_,
        the 14 standard fonts in PDF are limited to the ASCII-codepage. Therefore, in
        order to support other text in the exported PDF, one or more TTF font files have
        to be passed on to the exporting module.

        :returns: Additionl font settings for use in exporting PDFs.
        :rtype: :class:`PDFFontOptions` or :obj:`None <python:None>`
        """
        return self._pdf_font

    @pdf_font.setter
    @class_sensitive(PDFFontOptions)
    def pdf_font(self, value):
        self._pdf_font = value

    @property
    def print_max_width(self) -> Optional[int | float | Decimal]:
        """When printing the chart from the menu item in the burger menu, if the
        on-screen chart exceeds this width, it is resized. After printing or cancelled, it
        is restored.

        By default, set to ``780`` which makes
        the chart fit into typical paper format.

        .. note::

          This does not affect the chart when printing the web page as a whole.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._print_max_width

    @print_max_width.setter
    def print_max_width(self, value):
        self._print_max_width = validators.numeric(value, allow_empty = True)

    @property
    def scale(self) -> Optional[int | float | Decimal]:
        """Defines the scale or zoom factor for the exported image compared to the
        on-screen display. Defaults to ``2``.

        While for instance a 600px wide chart may look good on a website, it will look bad
        in print. The default scale of ``2`` makes this
        chart export to a 1200px PNG or JPG.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._scale

    @scale.setter
    def scale(self, value):
        self._scale = validators.numeric(value, allow_empty = True)

    @property
    def show_table(self) -> Optional[bool]:
        """If ``True``, shows an HTML table below the chart with the chart's current data.
        Defaults to ``False``.

        :returns: Flag indicating whether to display an HTML table with the export.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_table

    @show_table.setter
    def show_table(self, value):
        if value is None:
            self._show_table = None
        else:
            self._show_table = bool(value)

    @property
    def source_height(self) -> Optional[int | float | Decimal]:
        """The height of the original chart when exported, unless an explicit (JavaScript)
        ``chart.height`` is set, or a pixel width is set on the container.

        The height exported raster image is then multiplied by :meth:`Exporting.scale`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._source_height

    @source_height.setter
    def source_height(self, value):
        self._source_height = validators.numeric(value, allow_empty = True)

    @property
    def source_width(self) -> Optional[int | float | Decimal]:
        """The width of the original chart when exported, unless an explicit (JavaScript)
        ``chart.width`` is set, or a pixel width is set on the container.

        The width exported raster image is then multiplied by :meth:`Exporting.scale`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._source_width

    @source_width.setter
    def source_width(self, value):
        self._source_width = validators.numeric(value, allow_empty = True)

    @property
    def table_caption(self) -> Optional[bool | str]:
        """Caption for the data table. If not specified (:obj:`None <python:None>`)`), will
        default to the chart title.

        Also accepts a :class:`bool <python:bool>` value of ``False``, which disables
        the table caption entirely.

        :returns: Flag (value of ``False``) indicating whether to disable the table
          caption, or the text of the table caption to use if different from the chart
          title.
        :rtype: :class:`bool <python:bool>` or :class:`str <python:str>`, or
          :obj:`None <python:None>`
        """
        return self._table_caption

    @table_caption.setter
    def table_caption(self, value):
        if value is False:
            self._table_caption = False
        elif not value:
            self._table_caption = None
        else:
            self._table_caption = validators.string(value, allow_empty = False)

    @property
    def type(self) -> Optional[str]:
        """Default MIME type for exporting if the JavaScript ``chart.exportChart()`` is
        called without specifying a ``type`` option.

        Accepts:

          * ``'image/png'``
          * ``'image/jpeg'``
          * ``'application/pdf'``
          * ``'image/svg+xml'``

        Defaults to ``'image/png'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._type

    @type.setter
    def type(self, value):
        if not value:
            self._type = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['image/png',
                             'image/jpeg',
                             'application/pdf',
                             'image/svg+xml']:
                raise errors.HighchartsValueError(f'type expects a supported export MIME '
                                                  f'type. Received: "{value}"')

            self._type = value

    @property
    def url(self) -> Optional[str]:
        """The URL for the server module converting the SVG string to an image format. By
        default this points to Highchart's free web service:
        ``'{constants.DEFAULT_EXPORTING_URL}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises ValueError: if not a well-formed URL or path
        """
        return self._url

    @url.setter
    def url(self, value):
        try:
            self._url = validators.url(value, allow_empty = True)
        except ValueError:
            self._url = validators.path(value)

    @property
    def use_multi_level_headers(self) -> Optional[bool]:
        """If ``True``, uses multi-level (nested) headers in the exported data table.
        Defaults to ``True``.

        .. warning::

          If
          :meth:`Exporting.csv.column_header_formatter <ExportingCSV.column_header_formatter>`
          is specified, then the formatter must return objects for multi-level headers to
          work properly.

        :returns: Flag indicating whether to use multi-level headers in the data table.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._use_multi_level_headers

    @use_multi_level_headers.setter
    def use_multi_level_headers(self, value):
        if value is None:
            self._use_multi_level_headers = None
        else:
            self._use_multi_level_headers = bool(value)

    @property
    def use_rowspan_headers(self) -> Optional[bool]:
        """If ``True`` and using multi-level headers, uses rowspans in the data table for
        headers that only have one level. Defaults to ``True``.

        :returns: Flag indicating whether to use rowspans for single-level headers in a
          data table using multi-level headers.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._use_rowspan_headers

    @use_rowspan_headers.setter
    def use_rowspan_headers(self, value):
        if value is None:
            self._use_rowspan_headers = None
        else:
            self._use_rowspan_headers = bool(value)

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """An explicitly set pixel width for charts exported to PNG or JPG. Defaults to
        :obj:`None <python:None>`.

        .. note::

          If not specified (:obj:`None <python:None>`), the default pixel
          width is a function of the :meth:`Chart.width` or :meth:`Exporting.source_width`
          and :meth:`Exporting.scale`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_html': as_dict.get('allowHTML', None),
            'buttons': as_dict.get('buttons', None),
            'chart_options': as_dict.get('chartOptions', None),
            'csv': as_dict.get('csv', None),
            'enabled': as_dict.get('enabled', None),
            'error': as_dict.get('error', None),
            'fallback_to_export_server': as_dict.get('fallbackToExportServer', None),
            'filename': as_dict.get('filename', None),
            'form_attributes': as_dict.get('formAttributes', None),
            'lib_url': as_dict.get('libURL', None),
            'menu_item_definitions': as_dict.get('menuItemDefinitions', None),
            'pdf_font': as_dict.get('pdfFont', None),
            'print_max_width': as_dict.get('printMaxWidth', None),
            'scale': as_dict.get('scale', None),
            'show_table': as_dict.get('showTable', None),
            'source_height': as_dict.get('sourceHeight', None),
            'source_width': as_dict.get('sourceWidth', None),
            'table_caption': as_dict.get('tableCaption', None),
            'type': as_dict.get('type', None),
            'url': as_dict.get('url', None),
            'use_multi_level_headers': as_dict.get('useMultiLevelHeaders', None),
            'use_rowspan_headers': as_dict.get('useRowspanHeaders', None),
            'width': as_dict.get('width', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'accessibility': self.accessibility,
            'allowHTML': self.allow_html,
            'buttons': self.buttons,
            'chartOptions': self.chart_options,
            'csv': self.csv,
            'enabled': self.enabled,
            'error': self.error,
            'fallbackToExportServer': self.fallback_to_export_server,
            'filename': self.filename,
            'formAttributes': self.form_attributes,
            'libURL': self.lib_url,
            'menuItemDefinitions': self.menu_item_definitions,
            'pdfFont': self.pdf_font,
            'printMaxWidth': self.print_max_width,
            'scale': self.scale,
            'showTable': self.show_table,
            'sourceHeight': self.source_height,
            'sourceWidth': self.source_width,
            'tableCaption': self.table_caption,
            'type': self.type,
            'url': self.url,
            'useMultiLevelHeaders': self.use_multi_level_headers,
            'useRowspanHeaders': self.use_rowspan_headers,
            'width': self.width
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors, constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.data_labels import DataLabel
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.states import States


class BaseLevelOptions(HighchartsMeta):
    """Base class from which other Level Options classes inherit."""

    def __init__(self, **kwargs):
        self._border_color = None
        self._border_width = None
        self._data_labels = None
        self._level = None

        self.border_color = kwargs.get('border_color', None)
        self.border_width = kwargs.get('border_width', None)
        self.data_labels = kwargs.get('data_labels', None)
        self.level = kwargs.get('level', None)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the border surrounding each column or bar. Defaults to
        ``'#ffffff'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The width of the border surrounding each column or bar. If
        :obj:`None <python:None>`, defaults to ``1`` when there is room for a border, but
        to ``0`` when the columns are so dense that a border would cover the next column.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def data_labels(self) -> Optional[DataLabel]:
        """Options for the series data labels, appearing next to each data point.

        :rtype: :class:`DataLabel` or :obj:`None <python:None>`
        """
        return self._data_labels

    @data_labels.setter
    @class_sensitive(DataLabel)
    def data_labels(self, value):
        self._data_labels = value

    @property
    def level(self) -> Optional[int]:
        """Decides which level takes effect from the options set in the levels object.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._level

    @level.setter
    def level(self, value):
        self._level = validators.integer(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'data_labels': as_dict.get('dataLabels', None),
            'level': as_dict.get('level', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderColor': self.border_color,
            'borderWidth': self.border_width,
            'dataLabels': self.data_labels,
            'level': self.level
        }

        return untrimmed


class LevelOptions(BaseLevelOptions):
    """Set options on specific levels. Takes precedence over series options, but not
    node and link options."""

    def __init__(self, **kwargs):
        self._color_by_point = None
        self._link_opacity = None
        self._states = None

        self.color_by_point = kwargs.get('color_by_point', None)
        self.link_opacity = kwargs.get('link_opacity', None)
        self.states = kwargs.get('states', None)

        super().__init__(**kwargs)

    @property
    def color_by_point(self) -> Optional[bool]:
        """When using automatic point colors pulled from the global colors or
        series-specific collections, this option determines whether the chart should
        receive one color per series (``False``) or one color per point (``True``).

        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._color_by_point

    @color_by_point.setter
    def color_by_point(self, value):
        if value is None:
            self._color_by_point = None
        else:
            self._color_by_point = bool(value)

    @property
    def link_opacity(self) -> Optional[int | float | Decimal]:
        """Opacity for the links between nodes in sankey or similar diagrams. Defaults to
        ``0.5``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._link_opacity

    @link_opacity.setter
    def link_opacity(self, value):
        self._link_opacity = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def states(self) -> Optional[States]:
        """Configuration for state-specific configuration to apply to the data series.

        :rtype: :class:`States` or :obj:`None <python:None>`
        """
        return self._states

    @states.setter
    @class_sensitive(States)
    def states(self, value):
        self._states = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'data_labels': as_dict.get('dataLabels', None),
            'level': as_dict.get('level', None),

            'color_by_point': as_dict.get('colorByPoint', None),
            'link_opacity': as_dict.get('linkOpacity', None),
            'states': as_dict.get('states', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'colorByPoint': self.color_by_point,
            'dataLabels': self.data_labels,
            'linkOpacity': self.link_opacity,
            'states': self.states
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class ColorVariation(HighchartsMeta):
    """Configuration for a color variation to apply to all points on a level."""

    def __init__(self, **kwargs):
        self._key = None
        self._to = None

        self.key = kwargs.get('key', None)
        self.to = kwargs.get('to', None)

    @property
    def key(self) -> Optional[str]:
        """The key of a color variation. Currently only supports ``'brightness'``.
        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._key

    @key.setter
    def key(self, value):
        self._key = validators.string(value, allow_empty = True)

    @property
    def to(self) -> Optional[int | float | Decimal]:
        """The ending value of a color variation. The last sibling will receive this
        value. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._to

    @to.setter
    def to(self, value):
        self._to = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'key': as_dict.get('key', None),
            'to': as_dict.get('to', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'key': self.key,
            'to': self.to
        }

        return untrimmed


class LevelSize(HighchartsMeta):
    """Determines the width of the ring per level."""

    def __init__(self, **kwargs):
        self._unit = None
        self._value = None

        self.unit = kwargs.get('unit', None)
        self.value = kwargs.get('value', None)

    @property
    def unit(self) -> Optional[str]:
        """Indication of how to interpret :meth:`LevelSize.value`. Defaults to
        ``'weight'``.

        Accepts the following options:

          * ``'percentage'`` - gives a width relative to result of outer radius minus
            inner radius
          * ``'pixels'`` - gives the ring a fixed width in pixels
          * ``'weight'`` - takes the remaining width after percentage and pixels, and
            distributes it accross all "weighted" levels. The value relative to the sum of
            all weights determines the width.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._unit

    @unit.setter
    def unit(self, value):
        if not value:
            self._unit = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['percentage', 'pixels', 'weight']:
                raise errors.HighchartsValueError(f'unit expects "weight", "pixels", or '
                                                  f'"percentage". Received: {value}')
            self._unit = value

    @property
    def value(self) -> Optional[int | float | Decimal]:
        """The value used for calculating the width of the ring. Defaults to ``1``.

        .. note::

          The interpretation of this value is determined by :meth:`LevelSize.unit`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._value

    @value.setter
    def value(self, value_):
        self._value = validators.numeric(value_, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'unit': as_dict.get('unit', None),
            'value': as_dict.get('value', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'unit': self.unit,
            'value': self.value
        }

        return untrimmed


class SunburstLevelOptions(BaseLevelOptions):
    """Set options on specific levels for Sunburst Charts. Takes precedence over series
    options, but not node and link options."""

    def __init__(self, **kwargs):
        self._border_dash_style = None
        self._color = None
        self._color_variation = None
        self._level_size = None

        self.border_dash_style = kwargs.get('border_dash_style', None)
        self.color = kwargs.get('color', None)
        self.color_variation = kwargs.get('color_variation', None)
        self.level_size = kwargs.get('level_size', None)

        super().__init__(**kwargs)

    @property
    def border_dash_style(self) -> Optional[str]:
        """The dash style applied to all points which lie on the same level. Defaults to
        :obj:`None <python:None>`.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_dash_style

    @border_dash_style.setter
    def border_dash_style(self, value):
        if not value:
            self._border_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'border_dash_style expects a '
                                                  f'recognized value, but received: '
                                                  f'{value}')
            self._border_dash_style = value

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """Set a color on all points which lies on the same level. Defaults to
        :obj:`None <python:None>`.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`str <python:str>`
        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def color_variation(self) -> Optional[ColorVariation]:
        """Set a color variation on all points which lie on the same level. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`ColorVariation` or :obj:`None <python:None>`
        """
        return self._color_variation

    @color_variation.setter
    @class_sensitive(ColorVariation)
    def color_variation(self, value):
        self._color_variation = value

    @property
    def level_size(self) -> Optional[LevelSize]:
        """Set the width of the ring for all points which lie on the same level. Defaults
        to :obj:`None <python:None>`.

        :rtype: :class:`LevelSize` or :obj:`None <python:None>`
        """
        return self._level_size

    @level_size.setter
    @class_sensitive(LevelSize)
    def level_size(self, value):
        self._level_size = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'data_labels': as_dict.get('dataLabels', None),
            'level': as_dict.get('level', None),

            'border_dash_style': as_dict.get('borderDashStyle', None),
            'color': as_dict.get('color', None),
            'color_variation': as_dict.get('colorVariation', None),
            'level_size': as_dict.get('levelSize', None)
         }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderDashStyle': self.border_dash_style,
            'color': self.color,
            'colorVariation': self.color_variation,
            'levelSize': self.level_size
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class TreemapLevelOptions(BaseLevelOptions):
    """Set options on specific levels for Treemap Charts. Takes precedence over series
    options, but not node and link options."""

    def __init__(self, **kwargs):
        self._border_dash_style = None
        self._color = None
        self._color_variation = None
        self._layout_algoritm = None
        self._layout_starting_direction = None

        self.border_dash_style = kwargs.get('border_dash_style', None)
        self.color = kwargs.get('color', None)
        self.color_variation = kwargs.get('color_variation', None)
        self.layout_algorithm = kwargs.get('layout_algorithm', None)
        self.layout_starting_direction = kwargs.get('layout_starting_direction', None)

        super().__init__(**kwargs)

    @property
    def border_dash_style(self) -> Optional[str]:
        """The dash style applied to all points which lie on the same level. Defaults to
        :obj:`None <python:None>`.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_dash_style

    @border_dash_style.setter
    def border_dash_style(self, value):
        if not value:
            self._border_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'border_dash_style expects a '
                                                  f'recognized value, but received: '
                                                  f'{value}')
            self._border_dash_style = value

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """Set a color on all points which lies on the same level. Defaults to
        :obj:`None <python:None>`.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`str <python:str>`
        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def color_variation(self) -> Optional[ColorVariation]:
        """Set a color variation on all points which lie on the same level. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`ColorVariation` or :obj:`None <python:None>`
        """
        return self._color_variation

    @color_variation.setter
    @class_sensitive(ColorVariation)
    def color_variation(self, value):
        self._color_variation = value

    @property
    def layout_algorithm(self) -> Optional[str]:
        """This option decides which algorithm is used for setting position and dimensions
        of the points. Defaults to ``'sliceAndDice'``.

        Accepts the following (case-sensitive) values:

          * ``'sliceAndDice'``
          * ``'stripes'``
          * ``'squarified'``
          * ``'strip'``

        .. note::

          You also have the ability to extend Highcharts with your own layout algorithm.
          For more information, read the
          `original JavaScript documentation <https://www.highcharts.com/docs/chart-and-series-types/treemap#add-your-own-algorithm>`__.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._layout_algorithm

    @layout_algorithm.setter
    def layout_algorithm(self, value):
        self._layout_algorithm = validators.variable_name(value, allow_empty = True)

    @property
    def layout_starting_direction(self) -> Optional[str]:
        """Defines which direction the layout algorithm will start drawing. Defaults to
        ``'vertical'``.

        Accepts:

          * ``'vertical'``
          * ``'horizontal'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._layout_starting_direction

    @layout_starting_direction.setter
    def layout_starting_direction(self, value):
        if not value:
            self._layout_starting_direction = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['vertical', 'horizontal']:
                raise errors.HighchartsError(f'layout_starting_direction expects either '
                                             f'"vertical" or "horizontal". Received: '
                                             f'{value}')

            self._layout_starting_direction = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'data_labels': as_dict.get('dataLabels', None),
            'level': as_dict.get('level', None),

            'border_dash_style': as_dict.get('borderDashStyle', None),
            'color': as_dict.get('color', None),
            'color_variation': as_dict.get('colorVariation', None),

            'layout_algorithm': as_dict.get('layoutAlgorithm', None),
            'layout_starting_direction': as_dict.get('layoutStartingDirection', None)
         }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderDashStyle': self.border_dash_style,
            'color': self.color,
            'colorVariation': self.color_variation,
            'layoutAlgorithm': self.layout_algorithm,
            'layoutStartingDirection': self.layout_starting_direction
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



"""Unit tests for ``highcharts.utility_classes.javascript_functions``."""

import pytest

import os

from validator_collection import checkers
import esprima
from esprima.error_handler import Error as ParseError

from highcharts_core.utility_classes import javascript_functions as js_f
from highcharts_core import errors


def validate_js_function(as_str, _break_loop_on_failure = False, range = True):
    """Parse ``as_str`` as a valid JavaScript function.

    :param as_str: A putative JavaScript function definition
    :type as_str: :class:`str <python:str>`

    :returns: The Esprima AST, and a boolean flag to indicate whether it was prefixed
    :rtype: :class:`tuple <python:tuple>` of :class:`esprima.nodes.Script`,
      :class:`bool <python:bool>`
    """
    try:
        parsed = esprima.parseScript(as_str, loc = range, range = range)
    except ParseError:
        try:
            parsed = esprima.parseModule(as_str, loc = range, range = range)
        except ParseError as error:
            if not _break_loop_on_failure:
                as_str = f"""const testFunction = {as_str}"""
                return validate_js_function(as_str,
                                            _break_loop_on_failure = True,
                                            range = range)
            else:
                raise error

    return parsed, as_str


@pytest.mark.parametrize('kwargs, error', [
    ({}, None),
    ({
        'function_name': 'testFunction',
        'arguments': ['test1', 'test2'],
        'body': """return True;"""
    }, None),
    ({
        'function_name': 123
    }, TypeError),
    ({
        'arguments': 'not-a-list'
    }, TypeError),
    ({
        'body': 123
    }, TypeError),
])
def test_CallbackFunction__init__(kwargs, error):
    if not error:
        result = js_f.CallbackFunction(**kwargs)
        assert result is not None
        assert isinstance(result, js_f.CallbackFunction) is True
        for key in kwargs:
            assert kwargs[key] == getattr(result, key)

    else:
        with pytest.raises(error):
            result = js_f.CallbackFunction(**kwargs)


@pytest.mark.parametrize('as_dict, error', [
    ({}, None),
    ({
        'function_name': 'testFunction',
        'arguments': ['test1', 'test2'],
        'body': """return True;"""
    }, None),
    ({
        'function_name': 123
    }, TypeError),
    ({
        'arguments': 'not-a-list'
    }, TypeError),
    ({
        'body': 123
    }, TypeError),
])
def test_CallbackFunction_from_dict(as_dict, error):
    if not error:
        result = js_f.CallbackFunction.from_dict(as_dict)
        assert result is not None
        assert isinstance(result, js_f.CallbackFunction) is True
        for key in as_dict:
            assert as_dict[key] == getattr(result, key)

    else:
        with pytest.raises(error):
            result = js_f.CallbackFunction.from_dict(as_dict)


@pytest.mark.parametrize('kwargs, error', [
    ({}, None),
    ({
        'function_name': 'testFunction',
        'arguments': ['test1', 'test2'],
        'body': """return True;"""
    }, None),
])
def test_CallbackFunction__to_untrimmed_dict(kwargs, error):
    instance = js_f.CallbackFunction(**kwargs)
    if not error:
        result = instance._to_untrimmed_dict()
        assert result is not None
        assert isinstance(result, dict) is True
        for key in kwargs:
            assert kwargs[key] == result[key]

    else:
        with pytest.raises(error):
            result = instance._to_untrimmed_dict()


@pytest.mark.parametrize('kwargs, expected, error', [
    ({}, """function() {}""", None),
    ({
        'function_name': None,
        'arguments': ['test1', 'test2'],
        'body': """return True;"""
    }, """function(test1,test2) {\nreturn True;}""", None),
    ({
        'function_name': 'testFunction',
        'arguments': ['test1', 'test2'],
        'body': """return True;"""
    }, """function testFunction(test1,test2) {\nreturn True;}""", None),
])
def test_CallbackFunction__str__(kwargs, expected, error):
    instance = js_f.CallbackFunction(**kwargs)
    if not error:
        result = str(instance)
        assert result is not None
        assert isinstance(result, str) is True
        assert result == expected
        validate_js_function(result)
    else:
        with pytest.raises(error):
            result = str(instance)


@pytest.mark.parametrize('kwargs, filename, expected, error', [
    ({}, None, """function() {}""", None),
    ({
        'function_name': None,
        'arguments': ['test1', 'test2'],
        'body': """return True;"""
    }, None, """function(test1,test2) {\nreturn True;}""", None),
    ({
        'function_name': 'testFunction',
        'arguments': ['test1', 'test2'],
        'body': """return True;"""
    }, None, """function testFunction(test1,test2) {\nreturn True;}""", None),

    ({}, 'test1.js', """function() {}""", None),
    ({
        'function_name': None,
        'arguments': ['test1', 'test2'],
        'body': """return True;"""
    }, 'test2.js', """function(test1,test2) {\nreturn True;}""", None),
    ({
        'function_name': 'testFunction',
        'arguments': ['test1', 'test2'],
        'body': """return True;"""
    }, 'test3.js', """function testFunction(test1,test2) {\nreturn True;}""", None),
])
def test_CallbackFunction_to_js_literal(tmp_path, kwargs, filename, expected, error):
    instance = js_f.CallbackFunction(**kwargs)
    if filename:
        filename = os.path.join(tmp_path, filename)
    if not error:
        result = instance.to_js_literal(filename)
        assert result is not None
        assert isinstance(result, str) is True
        assert result == expected
        if filename:
            assert checkers.is_file(filename) is True
            with open(filename, 'r') as file_:
                result_as_str = file_.read()
                assert result_as_str == expected
    else:
        with pytest.raises(error):
            result = instance.to_js_literal(filename)


@pytest.mark.parametrize('original_str, error', [
    ("""function() {}""", None),
    ("""function(test1,test2) {\nreturn True;}""", None),
    ("""function testFunction(test1,test2) {\nreturn True;}""", None),
])
def test_CallbackFunction_convert_from_js_ast(original_str, error):
    original_parsed, updated_str = validate_js_function(original_str)
    unranged_result = validate_js_function(original_str, range = False)
    unranged_parsed = unranged_result[0]
    if original_parsed.body[0].type != 'FunctionDeclaration':
        property_definition = original_parsed.body[0].declarations[0].init
    else:
        property_definition = original_parsed.body[0]

    if not error:
        result = js_f.CallbackFunction._convert_from_js_ast(property_definition,
                                                            updated_str)
        assert result is not None
        assert isinstance(result, js_f.CallbackFunction) is True
        as_str = str(result)
        as_str_parsed, updated_as_str = validate_js_function(as_str, range = False)
        assert str(as_str_parsed) == str(unranged_parsed)
    else:
        with pytest.raises(error):
            result = js_f.CallbackFunction._convert_from_js_ast(property_definition,
                                                                updated_str)


@pytest.mark.parametrize('original_str, error', [
    ("""function() {}""", None),
    ("""function(test1,test2) {\nreturn true;}""", None),
    ("""function testFunction(test1,test2) {\nreturn true;}""", None),

    (123, TypeError),
    ("""const abc = 123;""", errors.HighchartsParseError),
])
def test_CallbackFunction_from_js_literal(original_str, error):
    if not error:
        unranged_result = js_f.CallbackFunction._validate_js_function(original_str,
                                                                      range = False)
        unranged_parsed = unranged_result[0]

        result = js_f.CallbackFunction.from_js_literal(original_str)
        assert result is not None
        assert isinstance(result, js_f.CallbackFunction) is True
        as_str = str(result)
        result_parsed = js_f.CallbackFunction._validate_js_function(as_str, range = False)
        as_str_parsed = result_parsed[0]
        assert str(as_str_parsed) == str(unranged_parsed)
    else:
        with pytest.raises(error):
            result = js_f.CallbackFunction.from_js_literal(original_str)


@pytest.mark.parametrize('kwargs, error', [
    ({}, None),
    ({'class_name': 'TestClass',
      'methods': []}, None),
    ({'class_name': 'TestClass',
      'methods': ["""function constructor() { return true; }"""]}, None),
    ({'class_name': 'TestClass',
      'methods': ["""constructor() { return true; }"""]}, None),
    ({'class_name': 'TestClass',
      'methods': ["""constructor() { return true; }""",
                  """testMethod(test1, test2) { return true; }"""]}, None),


    ({'methods': ["""function wrongName() { return true; }"""]}, errors.HighchartsJavaScriptError),
    ({'methods': ["""function() { return true;}"""]}, errors.HighchartsJavaScriptError),
])
def test_JavaScriptClass__init__(kwargs, error):
    if not error:
        result = js_f.JavaScriptClass(**kwargs)
        assert result is not None
        assert isinstance(result, js_f.JavaScriptClass)
        methods = result.methods or []
        if 'methods' in kwargs:
            method_strings = [x for x in kwargs['methods']]
            for key in kwargs:
                for method in methods:
                    for method_string in method_strings:
                        in_method_string = method.function_name in method_string
                        if in_method_string:
                            break

                    assert in_method_string is True

    else:
        with pytest.raises(error):
            result = js_f.JavaScriptClass(**kwargs)


@pytest.mark.parametrize('as_dict, error', [
    ({}, None),
    ({'className': 'TestClass',
      'methods': []}, None),
    ({'className': 'TestClass',
      'methods': ["""function constructor() { return true; }"""]}, None),
    ({'className': 'TestClass',
      'methods': ["""constructor() { return true; }"""]}, None),
    ({'className': 'TestClass',
      'methods': ["""constructor() { return true; }""",
                  """testMethod(test1, test2) { return true; }"""]}, None),


    ({'methods': ["""function wrongName() { return true; }"""]}, errors.HighchartsJavaScriptError),
    ({'methods': ["""function() { return true;}"""]}, errors.HighchartsJavaScriptError),
])
def test_JavaScriptClass_from_dict(as_dict, error):
    if not error:
        result = js_f.JavaScriptClass.from_dict(as_dict)
        assert result is not None
        assert isinstance(result, js_f.JavaScriptClass) is True
        if as_dict.get('className'):
            assert result.class_name is not None
            assert result.class_name == as_dict.get('className')
        methods = result.methods or []
        if 'methods' in as_dict:
            method_strings = [x for x in as_dict['methods']]
            for key in as_dict:
                for method in methods:
                    for method_string in method_strings:
                        in_method_string = method.function_name in method_string
                        if in_method_string:
                            break

                    assert in_method_string is True

    else:
        with pytest.raises(error):
            result = js_f.JavaScriptClass.from_dict(as_dict)


@pytest.mark.parametrize('kwargs, expected, error', [
    ({'class_name': 'TestClass',
      'methods': []},
     """class TestClass {\n}""",
     None),
    ({'class_name': 'TestClass',
      'methods': ["""function constructor() { return true; }"""]},
     """class TestClass {\nconstructor() {\n return true; \n}\n}""",
     None),
    ({'class_name': 'TestClass',
      'methods': ["""constructor(test1, test2) { return true; }"""]},
     """class TestClass {\nconstructor(test1,test2) {\n return true; \n}\n}""",
     None),
    ({'class_name': 'TestClass',
      'methods': ["""constructor() { return true; }""",
                  """testMethod(test1, test2) { return true; }"""]},
     """class TestClass {\nconstructor() {\n return true; \n}\ntestMethod(test1,test2) {\n return true; \n}\n}""",
     None),

    ({},
     """class None {\n}""",
     errors.HighchartsMissingClassNameError),
])
def test_JavaScriptClass__str__(kwargs, expected, error):
    instance = js_f.JavaScriptClass(**kwargs)
    if not error:
        result = str(instance)
        assert result is not None
        assert isinstance(result, str) is True
        print(result)
        print(expected)
        assert result == expected
        validate_js_function(result)
    else:
        with pytest.raises(error):
            result = str(instance)

@pytest.mark.parametrize('kwargs, expected, filename, error', [
    ({'class_name': 'TestClass',
      'methods': []},
     """class TestClass {\n}""",
     None,
     None),
    ({'class_name': 'TestClass',
      'methods': ["""function constructor() { return true; }"""]},
     """class TestClass {\nconstructor() {\n return true; \n}\n}""",
     None,
     None),
    ({'class_name': 'TestClass',
      'methods': ["""constructor(test1, test2) { return true; }"""]},
     """class TestClass {\nconstructor(test1,test2) {\n return true; \n}\n}""",
     None,
     None),
    ({'class_name': 'TestClass',
      'methods': ["""constructor() { return true; }""",
                  """testMethod(test1, test2) { return true; }"""]},
     """class TestClass {\nconstructor() {\n return true; \n}\ntestMethod(test1,test2) {\n return true; \n}\n}""",
     None,
     None),

    ({'class_name': 'TestClass',
      'methods': []},
     """class TestClass {\n}""",
     'test.js',
     None),
    ({'class_name': 'TestClass',
      'methods': ["""function constructor() { return true; }"""]},
     """class TestClass {\nconstructor() {\n return true; \n}\n}""",
     'test.js',
     None),
    ({'class_name': 'TestClass',
      'methods': ["""constructor(test1, test2) { return true; }"""]},
     """class TestClass {\nconstructor(test1,test2) {\n return true; \n}\n}""",
     'test.js',
     None),
    ({'class_name': 'TestClass',
      'methods': ["""constructor() { return true; }""",
                  """testMethod(test1, test2) { return true; }"""]},
     """class TestClass {\nconstructor() {\n return true; \n}\ntestMethod(test1,test2) {\n return true; \n}\n}""",
     'test.js',
     None),

    ({},
     """class None {\n}""",
     None,
     ValueError),
])
def test_JavaScriptClass_to_js_literal(tmp_path, kwargs, expected, filename, error):
    instance = js_f.JavaScriptClass(**kwargs)
    if filename:
        filename = os.path.join(tmp_path, filename)
    if not error:
        result = instance.to_js_literal(filename)
        assert result is not None
        assert isinstance(result, str) is True
        assert result == expected
        if filename:
            assert checkers.is_file(filename) is True
            with open(filename, 'r') as file_:
                result_as_str = file_.read()
                assert result_as_str == expected
    else:
        with pytest.raises(error):
            result = instance.to_js_literal(filename)


@pytest.mark.parametrize('original_str, error', [
    ("""class TestClass {\n}""", None),
    ("""class TestClass {\nconstructor() {\n return true; \n}\n}""", None),
    ("""class TestClass {\nconstructor(test1,test2) {\n return true; \n}\n}""", None),
    ("""class TestClass {\nconstructor() {\n return true; \n}\ntestMethod(test1,test2) {\n return true; \n}\n}""", None),

    ("""class None {\n}""", ValueError),
    ("""const notAClass = 123;""", errors.HighchartsParseError),
])
def test_JavaScriptClass_convert_from_js_ast(original_str, error):
    original_parsed, updated_str = validate_js_function(original_str)
    unranged_result = validate_js_function(original_str, range = False)
    unranged_parsed = unranged_result[0]
    definition = original_parsed.body[0]

    if not error:
        result = js_f.JavaScriptClass._convert_from_js_ast(definition, original_str)
        assert result is not None
        assert isinstance(result, js_f.JavaScriptClass) is True
        as_str = str(result)
        as_str_parsed, updated_as_str = validate_js_function(as_str, range = False)
        assert str(as_str_parsed) == str(unranged_parsed)
    else:
        with pytest.raises(error):
            result = js_f.JavaScriptClass._convert_from_js_ast(definition, original_str)


@pytest.mark.parametrize('original_str, error', [
    ("""class TestClass {\n}""", None),
    ("""class TestClass {\nconstructor() {\n return true; \n}\n}""", None),
    ("""class TestClass {\nconstructor(test1,test2) {\n return true; \n}\n}""", None),
    ("""class TestClass {\nconstructor() {\n return true; \n}\ntestMethod(test1,test2) {\n return true; \n}\n}""", None),

    ("""class None {\n}""", ValueError),
    ("""const notAClass = 123;""", errors.HighchartsParseError),
])
def test_JavaScriptClass_from_js_literal(original_str, error):
    if not error:
        unranged_result = validate_js_function(original_str, range = False)
        unranged_parsed = unranged_result[0]

        result = js_f.JavaScriptClass.from_js_literal(original_str)
        assert result is not None
        assert isinstance(result, js_f.JavaScriptClass) is True
        as_str = str(result)
        result_parsed = validate_js_function(as_str, range = False)
        as_str_parsed = result_parsed[0]
        assert str(as_str_parsed) == str(unranged_parsed)
    else:
        with pytest.raises(error):
            result = js_f.JavaScriptClass.from_js_literal(original_str)



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.states import States


class Marker(HighchartsMeta):
    """Options for the point markers of line-like series."""

    def __init__(self, **kwargs):
        self._enabled = None
        self._enabled_threshold = None
        self._fill_color = None
        self._height = None
        self._line_color = None
        self._line_width = None
        self._radius = None
        self._states = None
        self._symbol = None
        self._width = None

        self.enabled = kwargs.get('enabled', None)
        self.enabled_threshold = kwargs.get('enabled_threshold', None)
        self.fill_color = kwargs.get('fill_color', None)
        self.height = kwargs.get('height', None)
        self.line_color = kwargs.get('line_color', None)
        self.line_width = kwargs.get('line_width', None)
        self.radius = kwargs.get('radius', None)
        self.states = kwargs.get('states', None)
        self.symbol = kwargs.get('symbol', None)
        self.width = kwargs.get('width', None)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, markers are enabled. If ``False``, markers are disabled. If
        :obj:`None <python:None>`, markers are hidden when the data is dense and shown
        for more widespread data points.

        Defaults to :obj:`None <python:None>`

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def enabled_threshold(self) -> Optional[int]:
        """The threshold for how dense the point markers should be before they are hidden,
        assuming that :meth:`Marker.enabled` is :obj:`None <python:None>`. Defaults to
        ``2``.

        The number indicates the horizontal distance between the two closest points in the
        series, as multiples of the :Meth:`Marker.radius`. In other words, the default
        value of ``2`` means points are hidden if overlapping horizontally.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._enabled_threshold

    @enabled_threshold.setter
    def enabled_threshold(self, value):
        self._enabled_threshold = validators.integer(value,
                                                     allow_empty = True,
                                                     minimum = 0,
                                                     coerce_value = True)

    @property
    def fill_color(self) -> Optional[str | Gradient | Pattern]:
        """The fill color or gradient for the marker. Defaults to
        :obj:`None <python:None>`, which causes the series' or point's color to be used.

        :returns: The fill color for the marker.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._fill_color

    @fill_color.setter
    def fill_color(self, value):
        from highcharts_core import utility_functions
        self._fill_color = utility_functions.validate_color(value)

    @property
    def height(self) -> Optional[int | float | Decimal]:
        """Explicitly set the height of an image marker. Defaults to
        :obj:`None <python:None>`.

        .. warning::

          Only applies to image markers. When setting this property, :meth:`Marker.width`
          must also be set.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._height

    @height.setter
    def height(self, value):
        self._height = validators.numeric(value, allow_empty = True)

    @property
    def line_color(self) -> Optional[str | Gradient | Pattern]:
        """The line color or gradient for the marker's outline. Defaults to
        ``'#ffffff'``.

        :returns: The line color for the marker.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._line_color

    @line_color.setter
    def line_color(self, value):
        from highcharts_core import utility_functions
        self._line_color = utility_functions.validate_color(value)

    @property
    def line_width(self) -> Optional[int | float | Decimal]:
        """The line width (in pixels) applied to the marker's border. Defaults to
        ``0``.

        :returns: The line width to apply to the marker's border.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        self._line_width = validators.numeric(value, allow_empty = True)

    @property
    def radius(self) -> Optional[int | float | Decimal]:
        """The radius applied to the point marker. Defaults to
        ``4``.

        :returns: The radius to apply to the point marker.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = validators.numeric(value, allow_empty = True)

    @property
    def states(self) -> Optional[States]:
        """States for a single point marker.

        :rtype: :class:`States` or :obj:`None <python:None>`
        """
        return self._states

    @states.setter
    @class_sensitive(States)
    def states(self, value):
        self._states = value

    @property
    def symbol(self) -> Optional[str]:
        """A predefined shape or symbol for the marker. Defaults to
        :obj:`None <python:None>`.

        When :obj:`None <python:None>`, the symbol is pulled from (JavaScript)
        ``options.symbols``. Other possible values are:

          * ``'circle'``
          * ``'square'``
          * ``'diamond'``
          * ``'triangle'``
          * ``'triangle-down'``

        Additionally, the URL to a graphic can be given using the form:
        ``'url(graphic.png)'``.

        .. note::

          For an image marker to be applied to exported charts, its URL needs to be
          accessible by the export server.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._symbol

    @symbol.setter
    def symbol(self, value):
        if not value:
            self._symbol = None
        else:
            value = validators.string(value)
            value = value.lower()
            if not value.startswith('url') and value not in ['circle',
                                                             'square',
                                                             'diamond',
                                                             'triangle',
                                                             'triangle-down']:
                raise errors.HighchartsValueError(f'symbol must be a recognized value. '
                                                  f'Was: {value}')
            self._symbol = value

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """Explicitly set the width of an image marker. Defaults to
        :obj:`None <python:None>`.

        .. warning::

          Only applies to image markers. When setting this property, :meth:`Marker.height`
          must also be set.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'enabled': as_dict.get('enabled', None),
            'enabled_threshold': as_dict.get('enabledThreshold', None),
            'fill_color': as_dict.get('fillColor', None),
            'height': as_dict.get('height', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'radius': as_dict.get('radius', None),
            'states': as_dict.get('states', None),
            'symbol': as_dict.get('symbol', None),
            'width': as_dict.get('width', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'enabled': self.enabled,
            'enabledThreshold': self.enabled_threshold,
            'fillColor': self.fill_color,
            'height': self.height,
            'lineColor': self.line_color,
            'lineWidth': self.line_width,
            'radius': self.radius,
            'states': self.states,
            'symbol': self.symbol,
            'width': self.width
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.metaclasses import HighchartsMeta


class AnnotationPoint(HighchartsMeta):
    """Object representation of a shape point."""

    def __init__(self, **kwargs):
        self._x = None
        self._x_axis = None
        self._y = None
        self._y_axis = None

        self.x = kwargs.get('x', None)
        self.x_axis = kwargs.get('x_axis', None)
        self.y = kwargs.get('y', None)
        self.y_axis = kwargs.get('y_axis', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'annotations.shapes.point'

    @property
    def x(self) -> Optional[int | float | Decimal]:
        """The x position of the point.

        Units can be either in axis or chart pixel coordinates.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._x

    @x.setter
    def x(self, value):
        self._x = validators.numeric(value, allow_empty = True)

    @property
    def x_axis(self) -> Optional[str | int]:
        """This number defines which xAxis the point is connected to.

        It refers to either the axis id (as a :class:`str <python:str>`) or the index of
        the axis in the xAxis array (as an :class:`int <python:int>`).

        If the option is not configured or the axis is not found the point's x coordinate
        refers to the chart pixels.

        :rtype: :class:`str <python:str>` or :class:`int <python:int>`, or
          :obj:`None <python:None>`
        """
        return self._x_axis

    @x_axis.setter
    def x_axis(self, value):
        if value is None or value == '':
            self._x_axis = None
        else:
            try:
                try:
                    self._x_axis = validators.integer(value)
                except (ValueError, TypeError):
                    self._x_axis = validators.string(value)
            except ValueError:
                raise errors.HighchartsValueError('Unable to resolve value to a '
                                                  'supported type.')

    @property
    def y(self) -> Optional[int | float | Decimal]:
        """The y position of the point.

        Units can be either in axis or chart pixel coordinates.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._y

    @y.setter
    def y(self, value):
        self._y = validators.numeric(value, allow_empty = True)

    @property
    def y_axis(self) -> Optional[str | int]:
        """This number defines which yAxis the point is connected to.

        It refers to either the axis id (as a :class:`str <python:str>`) or the index of
        the axis in the yAxis array (as an :class:`int <python:int>`).

        If the option is not configured or the axis is not found the point's y coordinate
        refers to the chart pixels.

        :rtype: :class:`str <python:str>` or :class:`int <python:int>`, or
          :obj:`None <python:None>`
        """
        return self._y_axis

    @y_axis.setter
    def y_axis(self, value):
        if value is None or value == '':
            self._y_axis = None
        else:
            try:
                try:
                    self._y_axis = validators.integer(value)
                except (ValueError, TypeError):
                    self._y_axis = validators.string(value)
            except ValueError:
                raise errors.HighchartsValueError('Unable to resolve value to a '
                                                  'supported type.')

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'x': as_dict.get('x', None),
            'x_axis': as_dict.get('xAxis', None),
            'y': as_dict.get('y', None),
            'y_axis': as_dict.get('yAxis', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'x': self.x,
            'xAxis': self.x_axis,
            'y': self.y,
            'yAxis': self.y_axis
        }



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators, checkers

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.options.plot_options.generic import GenericTypeOptions
from highcharts_core.options.plot_options.levels import SunburstLevelOptions, LevelSize
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.breadcrumbs import BreadcrumbOptions
from highcharts_core.utility_classes.shadows import ShadowOptions
from highcharts_core.utility_classes.data_labels import SunburstDataLabel


class SunburstOptions(GenericTypeOptions):
    """General options to apply to all Sunburst series types.

    A Sunburst displays hierarchical data, where a level in the hierarchy is
    represented by a circle. The center represents the root node of the tree. The
    visualization bears a resemblance to both treemap and pie charts.

    .. figure:: ../../../_static/sunburst-example.png
      :alt: Sunburst Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._color_index = None
        self._crisp = None
        self._shadow = None
        self._allow_traversing_tree = None
        self._border_color = None
        self._border_width = None
        self._breadcrumbs = None
        self._center = None
        self._color_by_point = None
        self._fill_color = None
        self._level_is_constant = None
        self._levels = None
        self._level_size = None
        self._root_id = None
        self._size = None
        self._sliced_offset = None
        self._start_angle = None

        self.color_index = kwargs.get('color_index', None)
        self.crisp = kwargs.get('crisp', None)
        self.shadow = kwargs.get('shadow', None)
        self.allow_traversing_tree = kwargs.get('allow_traversing_tree', None)
        self.border_color = kwargs.get('border_color', None)
        self.border_width = kwargs.get('border_width', None)
        self.breadcrumbs = kwargs.get('breadcrumbs', None)
        self.center = kwargs.get('center', None)
        self.color_by_point = kwargs.get('color_by_point', None)
        self.fill_color = kwargs.get('fill_color', None)
        self.level_is_constant = kwargs.get('level_is_constant', None)
        self.levels = kwargs.get('levels', None)
        self.level_size = kwargs.get('level_size', None)
        self.root_id = kwargs.get('root_id', None)
        self.size = kwargs.get('size', None)
        self.sliced_offset = kwargs.get('sliced_offset', None)
        self.start_angle = kwargs.get('start_angle', None)

        super().__init__(**kwargs)

    @property
    def allow_traversing_tree(self) -> Optional[bool]:
        """If ``True``, the user can click on a point which is a parent and zoom in on its
        children. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_traversing_tree

    @allow_traversing_tree.setter
    def allow_traversing_tree(self, value):
        if value is None:
            self._allow_traversing_tree = None
        else:
            self._allow_traversing_tree = bool(value)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the border surrounding each slice. When :obj:`None <python:None>`,
        the border takes the same color as the slice fill. This can be used together with
        a :meth:`border_width <SunburstOptions.border_width>` to fill drawing gaps created by
        antialiazing artefacts in borderless pies. Defaults to ``'#ffffff'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The width of the border surrounding each slice. Defaults to ``1``.

        When setting the border width to ``0``, there may be small gaps between the slices
        due to SVG antialiasing artefacts. To work around this, keep the border width at
        ``0.5`` or ``1``, but set the :meth:`border_color <SunburstOptions.border_color>` to
        :obj:`None <python:None>` instead.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def breadcrumbs(self) -> Optional[BreadcrumbOptions]:
        """Options for the breadcrumbs, the navigation at the top leading the way up
        through the traversed levels. Defaults to :obj:`None <python:None>`.

        """
        return self._breadcrumbs

    @breadcrumbs.setter
    @class_sensitive(BreadcrumbOptions)
    def breadcrumbs(self, value):
        self._breadcrumbs = value

    @property
    def center(self) -> Optional[List[str | int | float | Decimal | constants.EnforcedNullType]]:
        """The center of the sunburst chart relative to the plot area.

        Can be percentages or pixel values. The default behaviour if
        :obj:`None <python:None>` is to center the pie so that all slices and data labels
        are within the plot area. As a consequence, the pie may actually jump around in a
        chart with dynamic values, as the data labels move. In that case, the center
        should be explicitly set, for example to ``["50%", "50%"]``.

        Defaults to ``['50%', '50%']``.

        :rtype: :obj:`None <python:None>` or :class:`list <python:list>` of numeric or
          :class:`str <python:str>` values
        """
        return self._center

    @center.setter
    def center(self, value):
        if not value:
            self._center = None
        else:
            value = validators.iterable(value)
            if len(value) != 2:
                raise errors.HighchartsValueError(f'center expects a 2-member array. '
                                                  f'Received a {len(value)}-member '
                                                  f'array.')
            processed_values = []
            for item in value:
                try:
                    item = validators.string(item)
                    if '%' not in item:
                        raise errors.HighchartsValueError('center expects either numbers '
                                                          'or percentage strings. No "%" '
                                                          'character found in string '
                                                          'received.')
                except TypeError:
                    item = validators.numeric(item)

                processed_values.append(item)

            self._center = processed_values

    @property
    def color_by_point(self) -> Optional[bool]:
        """When using automatic point colors pulled from the global colors or
        series-specific collections, this option determines whether the chart should
        receive one color per series (``False``) or one color per point (``True``).

        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._color_by_point

    @color_by_point.setter
    def color_by_point(self, value):
        if value is None:
            self._color_by_point = None
        else:
            self._color_by_point = bool(value)

    @property
    def color_index(self) -> Optional[int]:
        """When operating in :term:`styled mode`, a specific color index to use for the
        series, so that its graphic representations are given the class name
        ``highcharts-color-{n}``.

        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._color_index

    @color_index.setter
    def color_index(self, value):
        self._color_index = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def crisp(self) -> Optional[bool]:
        """If ``True``, each point or column edge is rounded to its nearest pixel in order
        to render sharp on screen. Defaults to ``True``.

        .. hint::

          In some cases, when there are a lot of densely packed columns, this leads to
          visible difference in column widths or distance between columns. In these cases,
          setting ``crisp`` to ``False`` may look better, even though each column is
          rendered blurry.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._crisp

    @crisp.setter
    def crisp(self, value):
        if value is None:
            self._crisp = None
        else:
            self._crisp = bool(value)

    @property
    def data_labels(self) -> Optional[SunburstDataLabel | List[SunburstDataLabel]]:
        """Options for the series data labels, appearing next to each data point.

        .. note::

          To have multiple data labels per data point, you can also supply a collection of
          :class:`DataLabel` configuration settings.

        :rtype: :class:`SunburstDataLabel <highcharts_core.utility_classes.data_labels.SunburstDataLabel>`, 
          :class:`list <python:list>` of 
            :class:`SunburstDataLabel <highcharts_core.utility_classes.data_labels.SunburstDataLabel>` or
            :obj:`None <python:None>`
        """
        return self._data_labels

    @data_labels.setter
    def data_labels(self, value):
        if not value:
            self._data_labels = None
        else:
            if checkers.is_iterable(value):
                self._data_labels = validate_types(value,
                                                   types = SunburstDataLabel,
                                                   allow_none = False,
                                                   force_iterable = True)
            else:
                self._data_labels = validate_types(value,
                                                   types = SunburstDataLabel,
                                                   allow_none = False)

    @property
    def fill_color(self) -> Optional[str | Gradient | Pattern]:
        """If the total sum of the pie's values is ``0``, the series is represented as an
        empty circle . The ``fill_color`` setting defines the color of that circle.
        Use :meth:`SunburstOptions.border_width` to set the border thickness.

        Defaults to :obj:`None <python:None>`.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, or :class:`Pattern`
        """
        return self._fill_color

    @fill_color.setter
    def fill_color(self, value):
        from highcharts_core import utility_functions
        self._fill_color = utility_functions.validate_color(value)

    @property
    def level_is_constant(self) -> Optional[bool]:
        """If ``True``, the level will be the same as the tree structure. If ``False``,
        the first level visible when drilling is considered to be level one. Defaults to
        ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._level_is_constant

    @level_is_constant.setter
    def level_is_constant(self, value):
        if value is None:
            self._level_is_constant = None
        else:
            self._level_is_constant = bool(value)

    @property
    def levels(self) -> Optional[List[SunburstLevelOptions]]:
        """Set options on specific levels. Takes precedence over series options, but not
        node and link options.

        :rtype: :obj:`None <python:None>`, or :class:`list <python:list>` of
          :class:`SunburstLevelOptions`
        """
        return self._levels

    @levels.setter
    @class_sensitive(SunburstLevelOptions, force_iterable = True)
    def levels(self, value):
        self._levels = value

    @property
    def level_size(self) -> Optional[LevelSize]:
        """Determines the width of the ring per level.

        :rtype: :class:`LevelSize` or :obj:`None <python:None>`
        """
        return self._level_size

    @level_size.setter
    @class_sensitive(LevelSize)
    def level_size(self, value):
        self._level_size = value

    @property
    def root_id(self) -> Optional[str]:
        """Indicates which point to use as a root in the visualization. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._root_id

    @root_id.setter
    def root_id(self, value):
        self._root_id = validators.string(value, allow_empty = True)

    @property
    def shadow(self) -> Optional[bool | ShadowOptions]:
        """Configuration for the shadow to apply to the tooltip. Defaults to
        ``False``.

        If ``False``, no shadow is applied.

        :returns: The shadow configuration to apply or a boolean setting which hides the
          shadow or displays the default shadow.
        :rtype: :class:`bool <python:bool>` or :class:`ShadowOptions`
        """
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        if isinstance(value, bool):
            self._shadow = value
        elif not value:
            self._shadow = None
        else:
            value = validate_types(value,
                                   types = ShadowOptions)
            self._shadow = value

    @property
    def size(self) -> Optional[str | int]:
        """The diameter of the pie relative to the plot area, expressed as a percentage or
        pixel value given as an integer.

        If :obj:`None <python:None>`, scales the pie to the plot area and gives room for
        data labels within the plot area.

        .. note::

          :meth:`SunburstOptions.sliced_offset` is also included in the default size
          calculation. As a consequence, the size of the pie may vary when points are
          updated and data labels more around. In that case it is best to set a fixed
          value, for example ``"75%"``.

        :rtype: :class:`str <python:str>`, :class:`int <python:int>`, or
          :obj:`None <python:None>`
        """
        return self._size

    @size.setter
    def size(self, value):
        if value is None:
            self._size = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError(f'size expects either a number or '
                                                      f'a % string. Received: {value}')
            except TypeError:
                value = validators.integer(value, minimum = 0)

            self._size = value

    @property
    def sliced_offset(self) -> Optional[int | float | Decimal]:
        """If a point is sliced, moved out from the center, how many pixels should it be
        moved? Defaults to ``10``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._sliced_offset

    @sliced_offset.setter
    def sliced_offset(self, value):
        self._sliced_offset = validators.numeric(value, allow_empty = True)

    @property
    def start_angle(self) -> Optional[int | float | Decimal]:
        """The start angle of the dependency wheel, in degrees where ``0`` is up. Defaults
        to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._start_angle

    @start_angle.setter
    def start_angle(self, value):
        self._start_angle = validators.numeric(value,
                                               allow_empty = True,
                                               minimum = 0,
                                               maximum = 360)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'allow_traversing_tree': as_dict.get('allowTraversingTree', None),
            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'breadcrumbs': as_dict.get('breadcrumbs', None),
            'center': as_dict.get('center', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'color_index': as_dict.get('colorIndex', None),
            'crisp': as_dict.get('crisp', None),
            'fill_color': as_dict.get('fillColor', None),
            'level_is_constant': as_dict.get('levelIsConstant', None),
            'levels': as_dict.get('levels', None),
            'level_size': as_dict.get('levelSize', None),
            'root_id': as_dict.get('rootId', None),
            'shadow': as_dict.get('shadow', None),
            'size': as_dict.get('size', None),
            'sliced_offset': as_dict.get('slicedOffset', None),
            'start_angle': as_dict.get('startAngle', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'allowTraversingTree': self.allow_traversing_tree,
            'borderColor': self.border_color,
            'borderWidth': self.border_width,
            'breadcrumbs': self.breadcrumbs,
            'center': self.center,
            'colorByPoint': self.color_by_point,
            'colorIndex': self.color_index,
            'crisp': self.crisp,
            'fillColor': self.fill_color,
            'levelIsConstant': self.level_is_constant,
            'levels': self.levels,
            'levelSize': self.level_size,
            'rootId': self.root_id,
            'shadow': self.shadow,
            'size': self.size,
            'slicedOffset': self.sliced_offset,
            'startAngle': self.start_angle
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class LegendLanguageOptions(HighchartsMeta):
    """Language options for the legend when used in accessibility mode."""

    def __init__(self, **kwargs):
        self._legend_item = None
        self._legend_label = None
        self._legend_label_no_title = None

        self.legend_item = kwargs.get('legend_item', None)
        self.legend_label = kwargs.get('legend_label', None)
        self.legend_label_no_title = kwargs.get('legend_label_no_title', None)

    @property
    def legend_item(self) -> Optional[str]:
        """Defaults to ``'Show {itemName}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._legend_item

    @legend_item.setter
    def legend_item(self, value):
        self._legend_item = validators.string(value, allow_empty = True)

    @property
    def legend_label(self) -> Optional[str]:
        """Defaults to ``'Chart legend: {legendTitle}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._legend_label

    @legend_label.setter
    def legend_label(self, value):
        self._legend_label = validators.string(value, allow_empty = True)

    @property
    def legend_label_no_title(self) -> Optional[str]:
        """Defaults to ``'Toggle series visibility, {chartTitle}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._legend_label_no_title

    @legend_label_no_title.setter
    def legend_label_no_title(self, value):
        self._legend_label_no_title = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'legend_item': as_dict.get('legendItem', None),
            'legend_label': as_dict.get('legendLabel', None),
            'legend_label_no_title': as_dict.get('legendLabelNoTitle', None),
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'legendItem': self.legend_item,
            'legendLabel': self.legend_label,
            'legendLabelNoTitle': self.legend_label_no_title
        }

        return untrimmed



"""Tests for ``highcharts.defs``."""

import pytest

from json.decoder import JSONDecodeError
from copy import deepcopy

from highcharts_core.options.defs import MarkerAttributeObject as cls
from highcharts_core.options.defs import MarkerASTNode as cls2
from highcharts_core.options.defs import MarkerDefinition as cls3

from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS_cls = [
    ({}, None),
    ({
      'id': 'test ID',
      'ref_x': 123,
      'ref_y': 456,
      'marker_height': 321
    }, None),
    ({
      'id': 'random ID',
      'new_key': 'test value',
      'hyphen-key': 123
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls)
def test_MarkerAttributeObject__init__(kwargs, error):
    kwargs_copy = deepcopy(kwargs)
    if not error:
        result = cls(**kwargs)
        print(kwargs_copy)
        assert result is not None
        assert isinstance(result, cls) is True
        for key in kwargs_copy:
            if isinstance(kwargs_copy[key], str) and kwargs[key].startswith('function'):
                continue
            if isinstance(kwargs_copy[key], str) and kwargs[key].startswith('class'):
                continue

            assert kwargs_copy[key] == result.get(key)
    else:
        with pytest.raises(error):
            result = cls(**kwargs)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls)
def test_MarkerAttributeObject__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls)
def test_MarkerAttributeObject_from_dict(kwargs, error):
    as_dict = to_js_dict(deepcopy(kwargs))

    if not error:
        instance = cls.from_dict(as_dict)
        assert instance is not None
        assert isinstance(instance, cls) is True
        for key in as_dict:
            if hasattr(instance, key):
                assert as_dict.get(key) == getattr(instance, key)
            else:
                assert as_dict.get(key) == instance.get(key)
    else:
        with pytest.raises(error):
            instance = cls.from_dict(as_dict)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls)
def test_MarkerAttributeObject_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


STANDARD_PARAMS_cls2 = [
    ({}, None),

]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls2)
def test_MarkerASTNode__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls2)
def test_MarkerASTNode__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls2)
def test_MarkerASTNode_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls2)
def test_MarkerASTNode_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


STANDARD_PARAMS_cls3 = [
    ({}, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls3)
def test_MarkerDefinition__init__(kwargs, error):
    kwargs_copy = deepcopy(kwargs)
    if not error:
        result = cls(**kwargs)
        print(kwargs_copy)
        assert result is not None
        assert isinstance(result, cls) is True
        for key in kwargs_copy:
            if isinstance(kwargs_copy[key], str) and kwargs[key].startswith('function'):
                continue
            if isinstance(kwargs_copy[key], str) and kwargs[key].startswith('class'):
                continue

            assert kwargs_copy[key] == result.get(key)
    else:
        with pytest.raises(error):
            result = cls(**kwargs)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls3)
def test_MarkerDefinition__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls3)
def test_MarkerDefinition_from_dict(kwargs, error):
    as_dict = to_js_dict(deepcopy(kwargs))

    if not error:
        instance = cls.from_dict(as_dict)
        assert instance is not None
        assert isinstance(instance, cls) is True
        for key in kwargs:
            assert kwargs.get(key) == instance.get(key)
    else:
        with pytest.raises(error):
            instance = cls.from_dict(as_dict)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls3)
def test_MarkerDefinition_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)



from typing import Optional, List

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta


class Binding(HighchartsMeta):
    """Basic interface for an event binding."""

    _class_name_default = None
    _max_steps = None

    def _private_setup(self, **kwargs):
        self._class_name_default = None
        self._max_steps = None

    def __init__(self, **kwargs):
        self._private_setup(**kwargs)

        self._init = None
        self._start = None
        self._steps = None
        self._end = None

        # Seems to be an out-dated functionality and is deprecated.
        # self._annotations_options = None

        self.class_name = kwargs.get('class_name', self._class_name_default)
        self.init = kwargs.get('init', None)
        self.start = kwargs.get('start', None)
        self.steps = kwargs.get('steps', None)
        self.end = kwargs.get('end', None)

        # Seems to be an out-dated functionality and is deprecated.
        # self.annotations_options = kwargs.get('annotations_options', None)

    @property
    def class_name(self) -> Optional[str]:
        """The classname used to bind events to.

        :rtype: :class:`str <python:str>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def init(self) -> Optional[str]:
        """The initial event, fired when the button is clicked.

        :rtype: :class:`str <python:str>`
        """
        return self._init

    @init.setter
    def init(self, value):
        self._init = validators.string(value, allow_empty = True)

    @property
    def start(self) -> Optional[str]:
        """Event fired on the first click on the first click.

        :rtype: :class:`str <python:str>`
        """
        return self._start

    @start.setter
    def start(self, value):
        self._start = validators.string(value, allow_empty = True)

    @property
    def steps(self) -> Optional[List[str]]:
        """Collection of sequential events, fired one after another on each of the user's
        clicks.

        .. note::

          The maximum number of events is controlled by a private property on the class,
          ``_max_steps``.

        :rtype: :class:`str <python:str>`
        """
        return self._steps

    @steps.setter
    def steps(self, value):
        if not value:
            self._steps = None
        else:
            value = validators.iterable(value, maximum_length = self._max_steps)
            self._steps = [validators.string(x) for x in value]

    @property
    def end(self) -> Optional[str]:
        """Last event to be called after the last step event.

        :rtype: :class:`str <python:str>`
        """
        return self._end

    @end.setter
    def end(self, value):
        self._end = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'class_name': as_dict.get('className', cls._class_name_default),
            'init': as_dict.get('init', None),
            'start': as_dict.get('start', None),
            'steps': as_dict.get('steps', None),
            'end': as_dict.get('end', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'className': self.class_name,
            'init': self.init,
            'start': self.start,
            'steps': self.steps,
            'end': self.end
        }

        return untrimmed


class CircleAnnotationBinding(Binding):
    """A circle annotation bindings. Includes ``start`` and one event in ``steps``
    array."""

    _class_name_default = 'highcharts-circle-annotation'
    _max_steps = 1

    def _private_setup(self, **kwargs):
        self._class_name_default = 'highcharts-circle-annotation'
        self._max_steps = 1

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class EllipseAnnotationBinding(Binding):
    """A ellipse annotation bindings. Includes ``start`` and two events in the
    ``steps`` array. The first event updates the second point, responsible for the rx
    width, and the second updates the ry width."""

    _class_name_default = 'highcharts-ellipse-annotation'
    _max_steps = 2

    def _private_setup(self, **kwargs):
        self._class_name_default = 'highcharts-ellipse-annotation'
        self._max_steps = 2

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class LabelAnnotationBinding(Binding):
    """A label annotation binding. Includes ``start`` event only."""

    _class_name_default = 'highcharts-label-annotation'
    _max_steps = 0

    def _private_setup(self, **kwargs):
        self._class_name_default = 'highcharts-label-annotation'
        self._max_steps = 0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class RectangleAnnotationBinding(Binding):
    """A rectangle annotation binding. Includes ``start`` and one event in the
    ``steps`` array."""

    _class_name_default = 'highcharts-rectangle-annotation'
    _max_steps = 1

    def _private_setup(self, **kwargs):
        self._class_name_default = 'highcharts-rectangle-annotation'
        self._max_steps = 1

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class Bindings(HighchartsMeta):
    """JavaScript event bindings for custom HTML buttons.

    Each JavaScript binding implements a simple event-driven interface:

      * ``class_name``: classname used to bind event to
      * ``init``: initial event, fired on button click
      * ``start``: fired on first click on a chart
      * ``steps``: array of sequential events fired one after another on each of users
        clicks
      * ``end``: last event to be called after last step event

    """

    def __init__(self, **kwargs):
        self._circle_annotation = None
        self._ellipse_annotation = None
        self._label_annotation = None
        self._rectangle_annotation = None

        self.circle_annotation = kwargs.get('circle_annotation', None)
        self.ellipse_annotation = kwargs.get('ellipse_annotation', None)
        self.label_annotation = kwargs.get('label_annotation', None)
        self.rectangle_annotation = kwargs.get('rectangle_annotation', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'navigation.bindings'

    @property
    def circle_annotation(self) -> Optional[CircleAnnotationBinding]:
        """A circle annotation bindings. Includes ``start`` and one event in ``steps``
        array.

        :rtype: :class:`CircleAnnotationBinding` or :obj:`None <python:None>`
        """
        return self._circle_annotation

    @circle_annotation.setter
    @class_sensitive(CircleAnnotationBinding)
    def circle_annotation(self, value):
        self._circle_annotation = value

    @property
    def ellipse_annotation(self) -> Optional[EllipseAnnotationBinding]:
        """A ellipse annotation bindings. Includes ``start`` and two events in the
        ``steps`` array. The first event updates the second point, responsible for the rx
        width, and the second updates the ry width.

        :rtype: :class:`EllipseAnnotationBinding` or :obj:`None <python:None>`
        """
        return self._ellipse_annotation

    @ellipse_annotation.setter
    @class_sensitive(EllipseAnnotationBinding)
    def ellipse_annotation(self, value):
        self._ellipse_annotation = value

    @property
    def label_annotation(self) -> Optional[LabelAnnotationBinding]:
        """A label annotation binding. Includes ``start`` event only.

        :rtype: :class:`LabelAnnotationBinding` or :obj:`None <python:None>`
        """
        return self._label_annotation

    @label_annotation.setter
    @class_sensitive(LabelAnnotationBinding)
    def label_annotation(self, value):
        self._label_annotation = value

    @property
    def rectangle_annotation(self) -> Optional[RectangleAnnotationBinding]:
        """A rectangle annotation bindings. Includes ``start`` and one event in the
        ``steps`` array.

        :rtype: :class:`RectangleAnnotationBinding` or :obj:`None <python:None>`
        """
        return self._rectangle_annotation

    @rectangle_annotation.setter
    @class_sensitive(RectangleAnnotationBinding)
    def rectangle_annotation(self, value):
        self._rectangle_annotation = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'circle_annotation': as_dict.get('circleAnnotation', None),
            'ellipse_annotation': as_dict.get('ellipseAnnotation', None),
            'label_annotation': as_dict.get('labelAnnotation', None),
            'rectangle_annotation': as_dict.get('rectangleAnnotation', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'circleAnnotation': self.circle_annotation,
            'ellipseAnnotation': self.ellipse_annotation,
            'labelAnnotation': self.label_annotation,
            'rectangleAnnotation': self.rectangle_annotation
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.axes.plot_bands import PlotBand as cls
from highcharts_core.options.axes.plot_bands import PlotLine as cls2
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'border_color': '#ccc',
      'border_width': 1,
      'class_name': 'some-class-name',
      'color': '#999',
      'events': {
        'click': """function(event) { return true; }""",
        'mousemove': """function(event) { return true; }""",
        'mouseout': """function(event) { return true; }""",
        'mouseover': """function(event) { return true; }"""
      },
      'from_': 123,
      'id': 'some-id',
      'label': {
          'align': 'center',
          'rotation': 0,
          'style': 'some-style-string',
          'text': 'The title of the plot band',
          'textAlign': 'left',
          'useHTML': False,
          'verticalAlign': 'middle',
          'x': -10,
          'y': 100
      },
      'to': 456,
      'z_index': 3
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_PlotBand__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_PlotBand__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_PlotBand_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_PlotBand_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('axes/plot_bands/01.js', False, None),

    ('axes/plot_bands/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('axes/plot_bands/01.js', True, None),

    ('axes/plot_bands/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_PlotBand_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)


STANDARD_PARAMS_2 = [
    ({}, None),
    ({
      'class_name': 'some-class-name',
      'color': '#fff',
      'dash_style': 'Dash',
      'events': {
        'click': """function(event) { return true; }""",
        'mousemove': """function(event) { return true; }""",
        'mouseout': """function(event) { return true; }""",
        'mouseover': """function(event) { return true; }"""
      },
      'label': {
          'align': 'center',
          'formatter': """function() { return true; }""",
          'rotation': 0,
          'style': 'some-style-string',
          'text': 'The title of the plot band',
          'textAlign': 'left',
          'useHTML': False,
          'verticalAlign': 'middle',
          'x': -10,
          'y': 100
      },
      'value': 123,
      'width': 2,
      'z_index': 3
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_PlotLine__init__(kwargs, error):
    Class__init__(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_PlotLine__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_PlotLine_from_dict(kwargs, error):
    Class_from_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_PlotLine_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('axes/plot_bands/02.js', False, None),

    ('axes/plot_bands/error-02.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('axes/plot_bands/02.js', True, None),

    ('axes/plot_bands/error-02.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_PlotLine_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls2, input_files, filename, as_file, error)



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import validate_types
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.animation import AnimationOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class AxisMarker(HighchartsMeta):
    """The triangular marker on a scalar color axis that points to the value of the
    hovered area."""

    def __init__(self, **kwargs):
        self._animation = None
        self._color = None
        self._width = None

        self.animation = kwargs.get('animation', None)
        self.color = kwargs.get('color', None)
        self.width = kwargs.get('width', None)

    @property
    def animation(self) -> Optional[bool | AnimationOptions]:
        """Animation for the marker as it moves between values. If
        :obj:`None <python:None>`, defaults to ``duration: 50``.

        .. hint::

          If set to ``False``, will disable animation of the marker.

        :rtype: :class:`AnimationOptions` or :class:`bool <python:bool>` or
          :obj:`None <python:None>`
        """
        return self._animation

    @animation.setter
    def animation(self, value):
        if value is False:
            self._animation = False
        else:
            self._animation = validate_types(value, AnimationOptions)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the marker. Defaults to :obj:`None <python:None>`, which
        applies '#999999'.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """The width of the marker. Defaults to :obj:`None <python:None>`, which applies
        a value of ``0.01``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'animation': as_dict.get('animation', None),
            'color': as_dict.get('color', None),
            'width': as_dict.get('width', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'animation': self.animation,
            'color': self.color,
            'width': self.width,
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.annotations import Annotation as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'animation': {
          'defer': 5
      },
      'control_point_options': {
          'positioner': """function(x, target) { return true; }"""
      },
      'crop': True,
      'draggable': 'xy',
      'events': {
          'add': """function(event) { return true; }""",
          'afterUpdate': """function(event) { return true; }""",
          'click': """function(event) { return true; }""",
          'remove': """function(event) { return true; }"""
      },
      'id': 'some-id',
      'label_options': {
          'accessibility': {
              'description': 'Description goes here'
          },
          'align': 'center',
          'allowOverlap': False,
          'backgroundColor': '#ccc',
          'borderColor': '#999',
          'borderRadius': 4,
          'borderWidth': 1,
          'className': 'some-class-name',
          'crop': False,
          'distance': 10,
          'format': 'format string',
          'formatter': """function() { return true; }""",
          'includeInDataExport': True,
          'overflow': 'justify',
          'padding': 12,
          'shadow': False,
          'shape': 'callout',
          'style': 'style-string-goes-here',
          'text': 'format string',
          'useHTML': False,
          'verticalAlign': 'top',
          'x': 5,
          'y': 10
      },
      'labels': [
          {
              'accessibility': {
                  'description': 'Description goes here'
              },
              'align': 'center',
              'allowOverlap': False,
              'backgroundColor': '#ccc',
              'borderColor': '#999',
              'borderRadius': 4,
              'borderWidth': 1,
              'className': 'some-class-name',
              'crop': False,
              'distance': 10,
              'format': 'format string',
              'formatter': """function() { return true; }""",
              'includeInDataExport': True,
              'overflow': 'justify',
              'padding': 12,
              'point': {
                  'x': 123,
                  'xAxis': 1,
                  'y': 456,
                  'yAxis': 1
              },
              'shadow': False,
              'shape': 'callout',
              'style': 'style-string-goes-here',
              'text': 'format string',
              'useHTML': False,
              'verticalAlign': 'top',
              'x': 5,
              'y': 10
          },
          {
              'accessibility': {
                  'description': 'Description goes here'
              },
              'align': 'center',
              'allowOverlap': False,
              'backgroundColor': '#ccc',
              'borderColor': '#999',
              'borderRadius': 4,
              'borderWidth': 1,
              'className': 'some-class-name',
              'crop': False,
              'distance': 10,
              'format': 'format string',
              'formatter': """function() { return true; }""",
              'includeInDataExport': True,
              'overflow': 'justify',
              'padding': 12,
              'point': 'point-id-goes-here',
              'shadow': False,
              'shape': 'callout',
              'style': 'style-string-goes-here',
              'text': 'format string',
              'useHTML': False,
              'verticalAlign': 'top',
              'x': 5,
              'y': 10
          },
          {
              'accessibility': {
                  'description': 'Description goes here'
              },
              'align': 'center',
              'allowOverlap': False,
              'backgroundColor': '#ccc',
              'borderColor': '#999',
              'borderRadius': 4,
              'borderWidth': 1,
              'className': 'some-class-name',
              'crop': False,
              'distance': 10,
              'format': 'format string',
              'formatter': """function() { return true; }""",
              'includeInDataExport': True,
              'overflow': 'justify',
              'padding': 12,
              'point': {
                  'x': 123,
                  'xAxis': 'axis-id',
                  'y': 456,
                  'yAxis': 'axis-id'
              },
              'shadow': False,
              'shape': 'callout',
              'style': 'style-string-goes-here',
              'text': 'format string',
              'useHTML': False,
              'verticalAlign': 'top',
              'x': 5,
              'y': 10
          }
      ],
      'shape_options': {
          'dashStyle': 'Solid',
          'fill': '#000',
          'height': 123,
          'r': 12,
          'ry': 24,
          'snap': 5,
          'src': 'https://www.somewhere.com',
          'stroke': '#ccc',
          'strokeWidth': 1,
          'type': 'rect',
          'width': 12,
          'xAxis': 1,
          'yAxis': 1
      },
      'shapes': [
          {
           'markerEnd': 'some-id-goes-here',
           'markerStart': 'some-id-goes-here',
           'point': {
               'x': 123,
               'xAxis': 1,
               'y': 456,
               'yAxis': 1
           },
           'points': [
               {
                'x': 123,
                'xAxis': 1,
                'y': 456,
                'yAxis': 1
               },
               'some-id-goes-here',
               """function() { return true; }"""
           ]
          }
      ],
      'visible': True,
      'z_index': 3
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('annotations/annotation/01.js', False, None),

    ('annotations/annotation/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('annotations/annotation/01.js', True, None),

    ('annotations/annotation/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional, List
from highcharts_core.options.series.funnel import FunnelSeries
from highcharts_core.options.series.pie import PieSeries
from highcharts_core.options.plot_options.pyramid import PyramidOptions, Pyramid3DOptions
from highcharts_core.options.series.data.single_point import SinglePointData
from highcharts_core.utility_functions import mro__to_untrimmed_dict


class PyramidSeries(FunnelSeries, PyramidOptions):
    """Options to configure a Pyramid series.

    A pyramid series is a special type of funnel, without neck and reversed by
    default.

    .. figure:: ../../../_static/pyramid-example.png
      :alt: Pyramid Example Chart
      :align: center

    """
    pass


class Pyramid3DSeries(PyramidOptions, PieSeries):
    """Options to configure a Pyramid 3D series.

    A pyramid 3d series is a special type of funnel, without neck and reversed by
    default.

    .. figure:: ../../../_static/pyramid_3d-example.png
      :alt: Pyramid 3D Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._gradient_for_sides = None

        self.gradient_for_sides = kwargs.get('gradient_for_sides', None)

        super().__init__(**kwargs)

    @property
    def data(self) -> Optional[List[SinglePointData]]:
        """Collection of data that represents the series. Defaults to
        :obj:`None <python:None>`.

        While the series type returns a collection of :class:`SinglePointData` instances,
        it accepts as input three different types of data:

        .. tabs::

          .. tab:: 1D Collection

            .. code-block::

              series = FunnelSeries()
              series.data = [0, 5, 3, 5]

            A one-dimensional collection of numerical values. Each member of the
            collection will be interpreted as a :meth:`y <SinglePointData.y>` value

          .. tab:: Object Collection

            A one-dimensional collection of :class:`SinglePointData` objects.

        :rtype: :class:`list <python:list>` of :class:`SinglePointData` or
          :obj:`None <python:None>`
        """
        return self._data

    @data.setter
    def data(self, value):
        if not value:
            self._data = None
        else:
            self._data = SinglePointData.from_array(value)

    @property
    def gradient_for_sides(self) -> Optional[bool]:
        """If ``True``, sides fill is set to a gradient. If ``False``, sides are rendered
        with a solid color. Defaults to :obj:`None <python:None>`, which is equivalent to
        ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`.
        """
        return self._gradient_for_sides

    @gradient_for_sides.setter
    def gradient_for_sides(self, value):
        if value is None:
            self._gradient_for_sides = None
        else:
            self._gradient_for_sides = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'point_description_format': as_dict.get('pointDescriptionFormat', None),
            'point_interval': as_dict.get('pointInterval', None),            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'data': as_dict.get('data', None),
            'id': as_dict.get('id', None),
            'index': as_dict.get('index', None),
            'legend_index': as_dict.get('legendIndex', None),
            'name': as_dict.get('name', None),
            'stack': as_dict.get('stack', None),
            'x_axis': as_dict.get('xAxis', None),
            'y_axis': as_dict.get('yAxis', None),
            'z_index': as_dict.get('zIndex', None),

            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'center': as_dict.get('center', None),
            'colors': as_dict.get('colors', None),
            'depth': as_dict.get('depth', None),
            'end_angle': as_dict.get('endAngle', None),
            'fill_color': as_dict.get('fillColor', None),
            'ignore_hidden_point': as_dict.get('ignoreHiddenPoint', None),
            'inner_size': as_dict.get('innerSize', None),
            'min_size': as_dict.get('minSize', None),
            'size': as_dict.get('size', None),
            'sliced_offset': as_dict.get('slicedOffset', None),
            'start_angle': as_dict.get('startAngle', None),
            'thickness': as_dict.get('thickness', None),

            'height': as_dict.get('height', None),
            'neck_height': as_dict.get('neckHeight', None),
            'neck_width': as_dict.get('neckWidth', None),
            'reversed': as_dict.get('reversed', None),
            'width': as_dict.get('width', None),

            'gradient_for_sides': as_dict.get('gradientForSides', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'data': self.data,
            'gradientForSides': self.gradient_for_sides,
            'colorAxis': self.color_axis,
            'colorIndex': self.color_index,
            'colorKey': self.color_key,
            'linecap': self.linecap
        }
        parent_as_dict = mro__to_untrimmed_dict(self, in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class Condition(HighchartsMeta):
    """Setting controls when to apply the
    :meth:`chart_options <ResponsiveRules.chart_options>`."""

    def __init__(self, **kwargs):
        self._callback = None
        self._max_height = None
        self._max_width = None
        self._min_height = None
        self._min_width = None

        self.callback = kwargs.get('callback', None)
        self.max_height = kwargs.get('max_height', None)
        self.max_width = kwargs.get('max_width', None)
        self.min_height = kwargs.get('min_height', None)
        self.min_width = kwargs.get('min_width', None)

    @property
    def callback(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function to gain complete control on when the responsive
        rule applies. Return ``true`` if it applies.

        .. hint::

          This feature enables you to use information besides the chart size to determine
          whether to enforce a set of :class:`ResponsiveRules` or not. Typical use cases
          might be to make the determination based on document size, available container
          size, or other other element on the page.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._callback

    @callback.setter
    @class_sensitive(CallbackFunction)
    def callback(self, value):
        self._callback = value

    @property
    def max_height(self) -> Optional[int | float | Decimal]:
        """The responsive rule applies if the chart height is less than this. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max_height

    @max_height.setter
    def max_height(self, value):
        if value is None:
            self._max_height = None
        else:
            self._max_height = validators.numeric(value,
                                                  allow_empty = True,
                                                  minimum = 0)

    @property
    def max_width(self) -> Optional[int | float | Decimal]:
        """The responsive rule applies if the chart width is less than this. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max_width

    @max_width.setter
    def max_width(self, value):
        if value is None:
            self._max_width = None
        else:
            self._max_width = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @property
    def min_height(self) -> Optional[int | float | Decimal]:
        """The responsive rule applies if the chart height is greater than this. Defaults
        to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min_height

    @min_height.setter
    def min_height(self, value):
        if value is None:
            self._min_height = None
        else:
            self._min_height = validators.numeric(value,
                                                  allow_empty = True,
                                                  minimum = 0)

    @property
    def min_width(self) -> Optional[int | float | Decimal]:
        """The responsive rule applies if the chart width is greater than this. Defaults
        to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min_width

    @min_width.setter
    def min_width(self, value):
        if value is None:
            self._min_width = None
        else:
            self._min_width = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'callback': as_dict.get('callback', None),
            'max_height': as_dict.get('maxHeight', None),
            'max_width': as_dict.get('maxWidth', None),
            'min_height': as_dict.get('minHeight', None),
            'min_width': as_dict.get('minWidth', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'callback': self.callback,
            'maxHeight': self.max_height,
            'maxWidth': self.max_width,
            'minHeight': self.min_height,
            'minWidth': self.min_width
        }

        return untrimmed


class ResponsiveRules(HighchartsMeta):
    """A set of rules for responsive settings."""

    def __init__(self, **kwargs):
        self._chart_options = None
        self._condition = None

        self.chart_options = kwargs.get('chart_options', None)
        self.condition = kwargs.get('condition', None)

    @property
    def chart_options(self) -> Optional[HighchartsMeta]:
        """A full set of chart :class:`HighchartsOptions` to apply as overrides to the general chart
        :class:`HighchartsOptions`. The chart options are applied when the given rule is active, as
        per :meth:`ResponsiveRules.condition`.

        .. note::

          A special case is configuration objects that take arrays, for example
          :meth:`Options.x_axis`, :meth:`Options.y_axis`, or :meth:`Options.series`. For
          these collections, an ``id`` option is used to map the new option set to an
          existing object. If an existing object of the same ``id`` is not found, the item
          of the same index is updated. So for example, setting :meth:`Options.series`
          with two series without an ``id``, will cause the existing chart's two series to
          be updated with respective options.

        :rtype: :class:`HighchartsOptions` or :obj:`None <python:None>`
        """
        return self._chart_options

    @chart_options.setter
    def chart_options(self, value):
        from highcharts_core.options import HighchartsOptions

        self._chart_options = validate_types(value, types = HighchartsOptions)

    @property
    def condition(self) -> Optional[Condition]:
        """Setting controls when to apply the
        :meth:`chart_options <ResponsiveRules.chart_options>`.

        :rtype: :class:`Condition` or :obj:`None <python:None>`
        """
        return self._condition

    @condition.setter
    @class_sensitive(Condition)
    def condition(self, value):
        self._condition = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'chart_options': as_dict.get('chartOptions', None),
            'condition': as_dict.get('condition', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'chartOptions': self.chart_options,
            'condition': self.condition
        }

        return untrimmed


class Responsive(HighchartsMeta):
    """Rules to apply for different screen or chart sizes.

    .. note::

      Each rule specifies additional chart options.

    """

    def __init__(self, **kwargs):
        self._rules = None

        self.rules = kwargs.get('rules', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'responsive'

    @property
    def rules(self) -> Optional[List[ResponsiveRules]]:
        """A set of rules for responsive settings.

        .. note::

          The rules are applied in the order provided.

        :rtype: :class:`list <python:list>` of :class:`ResponsiveRules` or
          :obj:`None <python:None>`
        """
        return self._rules

    @rules.setter
    @class_sensitive(ResponsiveRules, force_iterable = True)
    def rules(self, value):
        self._rules = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        return {
            'rules': as_dict.get('rules', None)
        }

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'rules': self.rules
        }

        return untrimmed
